"""
Buckets is a key-value interface to store and retrieve data objects up to 5 GB in size.

A full overview of the bucket interface may be found within PiCloud's
`documentation <http://docs.scivm.com/bucket.html>`_.

This module provides a python interface to your bucket.

In general, the various functions of this module use one or more of the following parameters:

* ``obj_path``: the key (path) of objects stored within the PiCloud bucket data store
* ``prefix``: Easy way to namespace objects; If present, the ``effective_obj_path`` is prefix + / + obj_path.
    If not present, the ``effective_obj_path`` is just obj_path.
    Referring to an object by obj_path="file", prefix="folder" is the same as
    obj_path="folder/file", prefix=""
* ``file_path``: Refers to a path on the local file system.

For convenience, in "copy" type functions (get, put, sync_to_cloud, etc.), the "destination" may be left
blank. If so, the destination will be set to os.path.basename(source).
(e.g. to upload, you need not put(file_path='foo', obj_path='foo'). put(file_path='foo') will suffice.)

.. note::
    Objects appearing in the bucket data store do **not** automatically appear on a
    job's mounted file system. Jobs **must** access bucket data with the get function.
"""

# TODO: Support files larger than 5 gb

from __future__ import with_statement
from __future__ import absolute_import
"""
Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2012 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see
http://www.gnu.org/licenses/lgpl-2.1.html
"""

"""Dev beware: list is defined in this module; list will not map to builtin list. Use builtin_list instead"""

import os
import base64
import sys
import logging
import time
import socket
import errno
import mimetypes
import random
import re
import urllib2
from __builtin__ import list as builtin_list
from ssl import SSLError
from itertools import islice

__httpConnection = None
__url = None

from .transport.adapter import SerializingAdapter
from .transport.network import HttpConnection
from .util import  min_args, max_args
from .util.zip_packer import Packer
from .cloud import CloudException, CloudTimeoutError
from scicloud import _getcloudnetconnection, _getcloud

cloudLog = logging.getLogger('Cloud.bucket')

S3_URL = 'https://s3.amazonaws.com/'
_bucket_new_query = 'bucket/new/'
_bucket_list_query = 'bucket/list/'
_bucket_get_query = 'bucket/get/'
_bucket_exists_query = 'bucket/exists/'
_bucket_info_query = 'bucket/info/'
_bucket_md5_query = 'bucket/md5/'
_bucket_remove_query = 'bucket/remove/'
_bucketmap_job_query = 'job/bucketmap/'

_bucket_make_public_query = 'bucket/make_public/'
_bucket_is_public_query = 'bucket/is_public/'
_bucket_make_private_query = 'bucket/make_private/'
_bucket_public_url_folder_query = 'bucket/public_url_folder/'

xml_chars = object # initialized on demand
"""
The functions can be viewed as functionally close to instance methods of cloud.Cloud
"""

def _post(conn, url, post_values, headers={}):
    """Use HttpConnection *conn* to issue a post request at *url* with values *post_values*"""

    #remove UNICODE from addresses
    url = url.decode('ascii', 'replace').encode('ascii', 'replace')

    if 'success_action_redirect' in headers:
        headers['success_action_redirect'] = headers['success_action_redirect'].decode('ascii', 'replace').encode('ascii', 'replace')
    if post_values and 'success_action_redirect' in post_values:
        post_values['success_action_redirect'] = post_values['success_action_redirect'].decode('ascii', 'replace').encode('ascii', 'replace')

    cloudLog.debug('post url %s with post_values=%s. headers=%s' % (url, post_values, headers))
    response =  conn.post(url, post_values, headers, use_gzip=False)

    return response

def _aws_retryable_post(conn, url, post_values, headers={}):
    """Wraps _post with ability to retry
    Sets some other necessary AWS settings"""
    retry_attempts = 8 # AWS can have scalability problems
    attempt = 0

    # likely not needed?
    #headers["Connection"] = "close"

    while attempt <= retry_attempts:
        try:
            return _post(conn, url, post_values, headers)
        except Exception, e:
            if isinstance(e, urllib2.HTTPError) and 200 < e.code < 300: # python 2.5 bug
                return e

            attempt += 1
            if attempt > retry_attempts:
                cloudLog.exception('_aws_retryable_post: Cannot connect to AWS')
                raise
            cloudLog.warn('_aws_retryable_post: Problem connecting to AWS. Retrying. \nError is %s' % str(e))
            c = attempt -1
            if (isinstance(e, socket.error) and getattr(e, 'errno', e.args[0]) == errno.ECONNREFUSED) or \
                (isinstance(e, urllib2.HTTPError) and e.code in [500, 503]):
                # guarantee at least a 3 second sleep on a connection refused error/500
                time.sleep(min(60, 3 + (1 << c) *random.random()))
            else:
                time.sleep(min(30, (1 << c) *random.random()))
            continue



class CloudBucketObject(object):
    """A CloudBucketObject provides a file-like interface to the contents of an object
    Seeking is not available."""

    __http_response = None

    __file_size = None
    __start_byte = None
    __end_byte = None
    __pos = None # current byte offset into file

    def __init__(self, action, ticket, file_size, start_byte=0, end_byte=None):
        """Wraps S3
        Action = url
        ticket = http headers
        """
        if not isinstance(start_byte, (int, long)):
            raise TypeError('start_byte must be an integer')

        if end_byte and not isinstance(end_byte, (int, long)):
            raise TypeError('end_byte must be an integer')

        self.__action = action
        self.__ticket = ticket
        self.__file_size = file_size
        self.__start_byte = start_byte
        self.__pos = self.__start_byte
        if end_byte:
            self.__end_byte = min(end_byte, file_size)
        else:
            self.__end_byte = file_size

        self.__connect()

    def __connect(self):
        """Connect to S3"""
        if self.__http_response:
            self.__http_response.close()
        self.__ticket['Range'] = 'bytes=%s-%s' % tuple(  [self.__pos, self.__end_byte]  )
        conn = _getcloudnetconnection()
        self.__http_response =  _aws_retryable_post(conn, self.__action, None, self.__ticket)

    def __protected_httpresp_op(self,op_name, *args, **kwargs):
        """Call HTTPResponse method name op_name in protected way
        On failure, reconnect
        """
        max_fails = 3
        for fail_num in xrange(max_fails):
            try:
                meth = getattr(self.__http_response, op_name)
#                if random.random() < 0.05:
#                    raise ssl.SSLError('hack')

                return meth(*args, **kwargs)

            except (SSLError, IOError):
                if fail_num == max_fails - 1:
                    raise
                cloudLog.info('Reconnecting to %s after failure %s at position %s',
                              self.__action, fail_num, self.__pos)

                time.sleep((1 << fail_num) * random.random())
                self.__connect()

    def __iter__(self):
        return self

    def close(self):
        """
        Close the object, blocking further I/O operations
        """
        return self.__http_response.close()

    @property
    def md5(self):
        """The md5 checksum of the contents
        Return None if not available"""
        md5str = self.__http_response.headers['etag'].strip('"')
        if '-' in md5str: # multipart; can't use
            return None
        else:
            return md5str

    def next(self):
        data = self.__protected_httpresp_op('next')
        self.__pos += len(data)
        return data

    def read(self, size=-1):
        """
        read([size]) -> read at most size bytes, returned as a string.

        If the size argument is negative or omitted, read until EOF is reached.
        """
        data = self.__protected_httpresp_op('read', size)
        self.__pos += len(data)
        return data

    def readline(self, size=-1):
        """
        readline([size]) -> next line from the file, as a string.

        Retain newline.  A non-negative size argument limits the maximum
        number of bytes to return (an incomplete line may be returned then).
        Return an empty string at EOF.
        """
        line = self.__protected_httpresp_op('readline', size)
        self.__pos += len(line)
        return line

    def readlines(self, sizehint=0):
        """
        readlines([size]) -> list of strings, each a line from the file.

        Call readline() repeatedly and return a list of the lines so read.
        The optional size argument, if given, is an approximate bound on the
        total number of bytes in the lines returned.
        """
        lines = self.__protected_httpresp_op('readlines', sizehint)
        self.__pos += sum((len(line) for line in lines))
        return lines

    def filesize(self):
        return self.__file_size

    def sizeofchunk(self):
        return self.__start_byte - self.__end_byte

    def tell(self):
        '''Returns current file position as an integer'''
        return self.__pos

    def end(self):
        return self.__end_byte

def _compute_md5(f, buffersize = 8192):
    """Computes the md5 hash of file-like object
    f must have seek ability to perform this operation
    buffersize controls how much of file is read at once

    Returns tuple of hexdigest, base64_encoded_md5_hash, size
    """

    try:
        start_loc = f.tell()
        f.seek(start_loc)  #test seek
    except (AttributeError, IOError), e:
        raise IOError('%s is not seekable. Cannot compute MD5 hash. Exact error is %s' % (str(f), str(e)))

    try:
        from hashlib import md5
    except ImportError:
        from md5 import md5

    m = md5()
    s = f.read(buffersize)
    while s:
        m.update(s)
        s = f.read(buffersize)

    size = f.tell() - start_loc
    f.seek(start_loc)
    hex_md5 = m.hexdigest()
    base64md5 = base64.encodestring(m.digest())
    return hex_md5, base64md5, size

def put(file_path, obj_path=None, prefix=None):
    """
    Transfer the file located on the local filesystem at ``file_path`` to the PiCloud bucket.

    The name of the uploaded object will be the ``effective_obj_path`` (see module help)
    If ``obj_path`` is omitted, the name will be the basename of ``file_path``.

    The uploaded object can be retrieved later with the get function.

    Example::

        cloud.bucket.put('data/names.txt')

    This will upload *data/names.txt* from the local filesystem to the PiCloud bucket
    and store the object with the effective_obj_path *names.txt*. It can be later retrieved
    via cloud.bucket.get('names.txt')"""

    if obj_path is None:
        obj_path = os.path.basename(file_path)
    elif not obj_path:
        raise ValueError('Cannot upload bucket object with obj_path "%s"' % obj_path)

    # open the requested file in binary mode (relevant in windows)
    content_type, content_encoding = mimetypes.guess_type(file_path)

    f = open(file_path, 'rb')
    _putf(f, obj_path, prefix, content_type, content_encoding)

def _validate_xml(s, entity_name):
    """Verify that s can be represented correctly in xml
    All key names must conform to XML 1.0
    """
    global xml_chars

    if not isinstance(s, unicode):
        try:
            uni_s = s.decode('utf-8')
        except UnicodeDecodeError:
            raise ValueError('Cannot represent %s in utf-8' % entity_name)
    else:
        uni_s = s


    if xml_chars == object: # regex not initialized
        try:
            xml_chars = re.compile(u'[^\x09\x0A\x0D\u0020-\uD7FF\uE000-\uFFFD\U00010000-\U0010FFFF]', re.U)
        except Exception: # narrow unicode python (http://wordaligned.org/articles/narrow-python)
            try:
                xml_chars = re.compile(u'[^\x09\x0A\x0D\u0020-\uD7FF\uE000-\uFFFD]', re.U)
            except Exception, e: # no idea?
                cloudLog.warning('Could not generate xml regex', exc_info = True)
                xml_chars = None

    elif xml_chars and xml_chars.search(uni_s):
        raise ValueError('%s contains characters illegal in XML 1.0' % entity_name)


def _get_effective_obj_path(obj_path, prefix):
    """Convert obj_path, prefix into an effective_obj_path
    Also, sanity check parameters
    """

    if not isinstance(obj_path, basestring):
        raise TypeError('obj_path must be a string')
    if '../..' in obj_path:
        raise ValueError('"../.." cannot be in obj_path')
    if obj_path.endswith('/'):
        raise ValueError('Cannot end a obj_path with a forward slash')
    if obj_path.startswith('/'):
        raise ValueError('Cannot begin a obj_path with a forward slash')
    if '//' in obj_path:
        raise ValueError('Cannot have consecutive forward slashes in obj_path')
    _validate_xml(obj_path, 'obj_path')

    if not prefix:
        return obj_path

    if hasattr(prefix, '__iter__'):
        prefix = '/'.join(prefix)

    if not isinstance(prefix, basestring):
        raise TypeError('prefix must be a string or list of strings')
    if '../..' in prefix:
        raise ValueError('"../.." cannot be in prefix')
    if prefix.startswith('/'):
        raise ValueError('Cannot begin a prefix with a forward slash')
    if '//' in prefix:
        raise ValueError('Cannot have consecutive forward slashes in prefix')
    _validate_xml(prefix, 'prefix')

    if not prefix.endswith('/'):
        prefix = prefix + '/'
    return prefix + obj_path


def _putf(f, obj_path, prefix=None, content_type=None, content_encoding=None):
    """
    helper for putf.
    Accepts arbitrary content_type and content_encoding
    """


    full_obj_path = _get_effective_obj_path(obj_path, prefix)

    fsize = 0 # file size. may not be computable

    if isinstance(f, basestring):
        from cStringIO import StringIO
        f = StringIO(f)

    conn = _getcloudnetconnection()

    try:
        #raise IOError
        hex_md5, content_md5, fsize = _compute_md5(f)

    except IOError:
        raise IOError('File object is not seekable. Cannot transmit')

    if fsize > 5000000000:
        raise ValueError('Cannot store bucket objects larger than 5GB on cloud.bucket')

    if fsize == 0:
        raise ValueError('Cannot store empty bucket objects')

    try:
        cloudLog.debug('bucket object obj_path in client: %s' % full_obj_path)
        # get a file ticket
        resp = conn.send_request(_bucket_new_query, {'name': full_obj_path,
                                                     'content-type' : content_type,
                                                     'content-encoding' : content_encoding,
                                                     'hex-md5' : hex_md5
                                                     })
        ticket = resp['ticket']
        params = resp['params']

        url = params['action']

        # update ticket
        ticket['file'] = f
        if content_md5:
            ticket['Content-MD5'] = content_md5

        resp =  _aws_retryable_post(conn, url, ticket)
        resp.read()

    finally:
        f.close()


def putf(f, obj_path, prefix=None):
    """
    Similar to put, but accepts any file-like object (file, StringIO, etc.) ``f``
    in lieu of a file_path.  The contents of the uploaded object will be
    f.read()

    The file-like object ``f`` must be seekable. Note that after putf returns, f
    will be closed
    """

    # infer mime types
    content_type, content_encoding = mimetypes.guess_type(obj_path)

    return _putf(f, obj_path, prefix, content_type, content_encoding)

def sync_to_cloud(file_path, obj_path=None, prefix=None):
    """
    Update bucket object if it has changed.

    put(file_path, obj_path, prefix) only if contents of ``file_path`` on local file system
    differ from bucket object ``effective_obj_path`` (or bucket object does not exist)
    """

    if not obj_path:
        obj_path = os.path.basename(file_path)

    # open the requested file in binary mode (relevant in windows)
    f = open(file_path, 'rb')
    local_md5, _, _ = _compute_md5(f)
    try:
        remote_md5 = _get_md5(obj_path, prefix, log_missing_file_error = False)
    except CloudException: #file not found
        remote_md5 = ''
    do_update = remote_md5 != local_md5

    cloudLog.debug('remote_md5=%s. local_md5=%s. uploading? %s',
                   remote_md5, local_md5, do_update
                   )
    if do_update:
        putf(f, obj_path, prefix)

class TruncatableList(builtin_list):
    truncated = False

def list(prefix=None, folderize=False, marker=None, max_keys=1000):
    """
    Retrieve obj_paths of all objects stored on PiCloud. Returns a list of keys in
    lexicographic order.

    * ``prefix``: Return only keys beginning with prefix.
    * ``folderize``: Treat listing as directory based; compact keys containing "/" into a single folder
        (a key is a folder if and only if it ends in "/")
        A folder can then be inspected by setting prefix equal to the folder name
    * marker: Return only keys with where key > marker
    * max_keys: Maximum number of keys that can be returned (max 1000). Fewer may be returned

    The list will have an attribute, *truncated*, that indicates if the listing is truncated.
    To see the next results, make a subsequent list query with marker set to list[-1]

    Use *iterlist* to avoid truncation
    """

    conn = _getcloudnetconnection()

    if max_keys > 1000:
        max_keys = 1000

    resp = conn.send_request(_bucket_list_query, {'prefix': prefix,
                                                  'delimiter': '/' if folderize else None,
                                                  'marker': marker,
                                                  'max_keys': max_keys})

    files = TruncatableList(resp['files'])
    truncated = resp['truncated']

    files.truncated = truncated
    return files

def iterlist(prefix=None, folderize=False, num_readahead_keys=1000):
    """Retrieve obj_paths of all objects stored on PiCloud. Returns an iterator that produces keys
    in lexicographic order.  Unlike list() this guarantees that all keys will be returned:

    * ``prefix``: Return only keys beginning with prefix.
    * ``folderize``: Treat listing as directory based; compact keys containing "/" into a single folder
        (a key is a folder if and only if it ends in "/")
        A folder can then be inspected by setting prefix equal to the folder name
    * ``num_readahead_keys``: Number of keys to prefetch from amazon

    .. warning:: Usage of iterlist is not recommended if result set will exceed 3,000 entries.
        Consider using prefixes and folderize to narrow results
    """

    def listgen():
        marker = None
        truncated = True
        while truncated:
            obj_name_chunk = list(prefix, folderize, marker, num_readahead_keys)
            truncated = obj_name_chunk.truncated
            for obj_name in obj_name_chunk:
                yield obj_name
            if not truncated:
                break
            marker = obj_name

    return listgen()

def exists(obj_path, prefix=None):
    """Return boolean indicating if PiCloud bucket object named ``effective_obj_path`` exists"""
    conn = _getcloudnetconnection()

    full_obj_path = _get_effective_obj_path(obj_path, prefix)
    resp = conn.send_request(_bucket_exists_query, {'name': full_obj_path})
    exists = resp['exists']
    return exists

def info(obj_path, prefix=None):
    """Return information about the PiCloud bucket object ``effective_obj_path``

    Information includes size, created time, last modified time, md5sum, public URL (if any),
    and any headers set with ``make_public``
    """

    conn = _getcloudnetconnection()

    full_obj_path = _get_effective_obj_path(obj_path, prefix)
    resp = conn.send_request(_bucket_info_query, {'name': full_obj_path})
    del resp['data']
    if 'url' in resp:
        resp['url'] = S3_URL+resp['url']

    return resp

def _get_md5(obj_path, prefix=None, log_missing_file_error = True):
    conn = _getcloudnetconnection()

    full_obj_path = _get_effective_obj_path(obj_path, prefix)
    resp = conn.send_request(_bucket_md5_query, {'name': full_obj_path},
                             log_cloud_excp = log_missing_file_error)
    md5sum = resp['md5sum']

    if '-' in md5sum: # multipart; can't rely on md5
        return None

    return md5sum

def get_md5(obj_path, prefix=None):
    """Return the md5 checksum of the PiCloud bucket object ``effective_obj_path``
    Return None if not available
    """
    return _get_md5(obj_path, prefix)

def remove(obj_paths, prefix=None):
    """Removes object(s) named ``effective_obj_paths`` from PiCloud bucket

    obj_paths can be a single object or a list of objects
    """
    conn = _getcloudnetconnection()

    if not hasattr(obj_paths, '__iter__'):
        obj_paths = [obj_paths]

    obj_paths_iter = obj_paths.__iter__()

    removed = False
    while True:
        paths_to_remove = builtin_list(islice(obj_paths_iter, 1000))
        if not paths_to_remove:
            break

        full_obj_paths =[ _get_effective_obj_path(obj_path, prefix) for obj_path in paths_to_remove]
        resp = conn.send_request(_bucket_remove_query, {'name': full_obj_paths})
        removed = resp['removed']

    return removed

def remove_prefix(prefix):
    """Removes all objects beginning with prefix"""
    objects_iter = iterlist(prefix)

    while True:
        batched_objects = builtin_list(islice(objects_iter, 1000))
        if not batched_objects:
            break

        remove(batched_objects)

def make_public(obj_path, prefix=None, headers={}, reset_headers = False):
    """Makes the PiCloud bucket object ``effective_obj_path`` publicly accessible by a URL
    Returns public URL

    Additionally, you can control the HTTP headers that will be in the response to a request
    for the URL with the ``headers`` dictionary.

    Possible standard HTTP headers are:

    * content-type
    * content-encoding
    * content-disposition
    * cache-control
    All other headers are considered custom and will have x-amz-meta- prepended to them.

    Example:
    make_public('foo',headers={'content-type': 'text/x-python', 'purpose' : 'basic_script'}
    might return \https://s3.amazonaws.com/pi-user-buckets/ddasy/foo

    The headers in the response to a request for \https://s3.amazonaws.com/pi-user-buckets/ddasy/foo
    will include:

    * content-type: text/x-python
    * x-amz-meta-purpose: basic_script

    Clear all custom headers, other than content-type and content-encoding, by
    setting ``reset_headers`` to True

    .. note:: Default content-type and content-encoding are inferred during the original
        cloud.bucket.put(..) call from the ``file_path`` and ``obj_path``.
    """
    conn = _getcloudnetconnection()

    full_obj_path = _get_effective_obj_path(obj_path, prefix)
    post_values = {'name' : full_obj_path,
                   'reset_headers' : reset_headers}
    for key, val in headers.items():
        try:
            post_values['bh_' + key] = val.decode('ascii').encode('ascii')
        except (UnicodeDecodeError, UnicodeEncodeError):
            raise TypeError('header values must be ASCII strings')

    resp = conn.send_request(_bucket_make_public_query, post_values)
    public_url = S3_URL+resp['url']
    return public_url

def public_url_folder():
    """Return HTTP path that begins all your public bucket URLs.
    e.g. object 'foo' (if is_public) will be found at
        public_url_folder() + foo
    """
    conn = _getcloudnetconnection()
    resp = conn.send_request(_bucket_public_url_folder_query, {})
    return S3_URL+resp['url']

def is_public(obj_path, prefix=None):
    """Determine if the PiCloud bucket object ``effective_obj_path``
    is publicly accessible by a URL

    Return public URL if it is; otherwise False
    """
    conn = _getcloudnetconnection()

    full_obj_path = _get_effective_obj_path(obj_path, prefix)
    resp = conn.send_request(_bucket_is_public_query, {'name': full_obj_path})
    if resp['status']:
        public_url = S3_URL+resp['url']
        return public_url
    else:
        return resp['status']


def make_private(obj_path, prefix=None):
    """Removes the public URL associated with the PiCloud bucket object ``effective_obj_path``"""
    conn = _getcloudnetconnection()

    full_obj_path = _get_effective_obj_path(obj_path, prefix)
    resp = conn.send_request(_bucket_make_private_query, {'name': full_obj_path})

def _ready_file_path(file_path, obj_path=None):
    """Ensure that the directory where file_path is pointed to exists (create it it doesn't)
    Return file_path (extract it from obj_path if it doesn't exist)
    """
    if not file_path:
        file_path = os.path.basename(obj_path)

    # create directory if needed
    dest_dir = os.path.dirname(file_path)
    if dest_dir:
        try:
            os.makedirs(dest_dir)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise

    return file_path


def get(obj_path, file_path=None, prefix=None, start_byte=0, end_byte=None, _retries=1):
    """
    Retrieve the bucket object referenced by ``effective_obj_path`` from PiCloud and
    save it to the local file system at ``file_path``.

    Example::

        cloud.bucket.get('names.txt','data/names.txt')

    This will retrieve the *names.txt* bucket object from PiCloud and save it locally to
    *data/names.txt*.

    If ``file_path`` is None, it will be set to the basename of ``obj_path``.

    An optional byte_range can be specified using ``start_byte`` and ``end_byte``,
    where only the data between ``start_byte`` and ``end_byte`` is returned.

    An ``end_byte`` of None or exceeding file size is interpreted as a request to retrieve to end of file.
    """

    file_path = _ready_file_path(file_path, obj_path)

    cloud_file = getf(obj_path, prefix, start_byte, end_byte)
    remote_md5 = cloud_file.md5

    chunk_size = 64000

    f = open(file_path, 'wb')

    while True:
        data = cloud_file.read(chunk_size)
        if not data:
            break
        f.write(data)

    f.close()

    # Only possible to validate MD5 for full file
    if remote_md5 and not start_byte and not end_byte:
        f = open(file_path, 'rb')
        hex_md5, _, _ = _compute_md5(f)
        f.close()

        if hex_md5 != remote_md5:
            msg = 'Local MD5 %s did not match remote MD5 %s for %s' % (hex_md5, remote_md5, file_path)
            cloudLog.error(msg)
            if _retries > 0:
                return get(obj_path, file_path, prefix, start_byte, end_byte, _retries - 1)

            raise RuntimeError(msg)


def sync_from_cloud(obj_path, file_path=None, prefix=None):
    """
    Download bucket object if it has changed.

    get(obj_path, file_path, prefix) only if contents of ``file_path`` on local file system
    differ from bucket object ``effective_obj_path`` (or local file does not exist)
    """

    file_path = _ready_file_path(file_path, obj_path)

    if not os.path.exists(file_path):
        do_update = True
    else:
        f = open(file_path, 'rb')
        local_md5, _, _ = _compute_md5(f)
        f.close()
        try:
            remote_md5 = _get_md5(obj_path, prefix, log_missing_file_error=False)
        except CloudException: #file not found
            remote_md5 = ''

        do_update = remote_md5 != local_md5
        cloudLog.debug('remote_md5=%s. local_md5=%s. downloading? %s',
                       remote_md5, local_md5, do_update)
    if do_update:
        get(obj_path, file_path, prefix)

def mpsafe_get(obj_path, file_path=None, prefix=None, start_byte=0, end_byte=None,
               timeout=None, do_sync=False):
    """Multiprocessing-safe variant of get.
    If do_sync is false:
        Download file if and only if file_path does not exist. (atomically checked)
    If do_sync is true:
        Download file if and only if copy at file_path does not match remote copy (atomically checked)
    Regardless, If another process is downloading, wait until process has finished

    If timeout is reached, CloudTimeoutException is raised
    """

    from .util.cloghandler import portalocker

    file_path = _ready_file_path(file_path, obj_path)

    # lock the file itself
    try:
        #fd = os.open(file_path, os.O_CREAT|os.O_RDONLY)
        fobj = open(file_path,'a')
    except OSError as oe:
        if oe.errno != errno.EEXIST:
            raise

    expire_time = timeout + time.time() if timeout else None
    while True:
        try:
            portalocker.lock(fobj, (portalocker.LOCK_EX | portalocker.LOCK_NB))
        except portalocker.LockException as le:
            if le.args[0] != portalocker.LockException.LOCK_FAILED:
                raise
            if expire_time and time.time() > expire_time:
                raise CloudTimeoutError('could not acquire file lock within timeout')

            time.sleep(0.2) # delay
        else:
            break

    # We now have the lock; determine what to do based on file size
    # Zero file size is special as buckets cannot contained zero sized files
    # If file currently is zero-sized, we are the first writer: download it
    #   Else, someone else got to it first
    fsize = os.path.getsize(file_path)
    if not fsize: # download
        get(obj_path, file_path, prefix, start_byte, end_byte)
    elif do_sync:
        sync_from_cloud(obj_path, file_path, prefix)

    fobj.close()  # releases lock


def getf(obj_path, prefix=None, start_byte=0, end_byte=None):
    """
    Retrieve the object referenced by ``effective_obj_path`` from PiCloud.
    Return value is a CloudBucketObject (file-like object) that can be read() to
    retrieve the object's contents

    An optional byte_range can be specified using ``start_byte`` and ``end_byte``,
    where only the data between ``start_byte`` and ``end_byte`` is returned and made
    accessible to the CloudBucketObject.  The returned CloudBucketObject.tell() will
    initialized to ``start_byte``.

    An ``end_byte`` of None or exceeding file size is interpreted as a request to retrieve to end of file.
    """

    full_obj_path = _get_effective_obj_path(obj_path, prefix)
    conn = _getcloudnetconnection()

    resp = conn.send_request(_bucket_get_query, {'name': full_obj_path})

    ticket = resp['ticket']
    params = resp['params']
    file_size = params['size']

    if not start_byte:
        start_byte = 0

    if file_size and (not end_byte or end_byte > file_size):
        end_byte = file_size

    cloud_file = CloudBucketObject( params['action'], ticket, file_size, start_byte, end_byte )

    return cloud_file




'''
Queue message attachment logic
'''

"""
Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2013 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html    
"""

import numbers
import os
import time
import threading
import thread
import sys
import traceback
from subprocess import Popen, PIPE

from ..cloud import CloudException
from ..queue import Retry, MaxRetryException
from inspect import getargspec

import logging
cloudLog = logging.getLogger('Cloud.queue')


def _handle_action_dct_excp(action_dct, excp, message, cur_retry, max_retries):
    """Push message (and relevant errors) to queue defined by action_dct
    Interprets Retry based on inner_excp
    """
    
    if isinstance(excp, Retry) and MaxRetryException not in action_dct:
        tb = excp.traceback_info
        test_excp = excp.inner_excp_value or excp
    else:
        tb = traceback.format_exc()
        test_excp = excp

    try:
        for parent_excp in test_excp.__class__.mro():
            if parent_excp in action_dct:
                action = action_dct[parent_excp]
    
                error_message = {'exception': excp,
                                 'traceback': traceback.format_exc(),
                                 'source_message': message,
                                 'cur_retry' : cur_retry}
    
                # TODO: Asynchronous?
                action['queue'].push([error_message], delay=action.get('delay', 0))
                break
    except Exception as e:
        print >> sys.stderr, 'Exception handler raised exception'
        traceback.print_exc()
        print >> sys.stderr, ''

def launch_attachment(input_queue, message_handler, expand_iterable_output, output_queues, 
                       batch_size, readers_per_job, retry_on=[], max_retries=None, retry_delay=None,
                       on_error={}, _job_shutdown_timeout=20):
    """Creates the function to be executed on PiCloud."""

    def attachment_runner(message_queue, result_queue):

        message_handler_func = message_handler if callable(message_handler) else message_handler.message_handler
        pre_handling_func = message_handler.pre_handling if callable(getattr(message_handler, 'pre_handling', None)) else None
        post_handling_func = message_handler.post_handling if callable(getattr(message_handler, 'post_handling', None)) else None
        sentinel = None
        
        message_handler_argspec = getargspec(message_handler_func)
        
        if pre_handling_func:
            try:
                pre_handling_func()
            except:
                sys.stderr.write('pre_handling raised exception.\n')
                traceback.print_exc()
                sys.stderr.write('\n')
                sentinel = Exception('pre_handling errored:\n' + traceback.format_exc())

        while True:
            
            encoded_message = message_queue.get()
            #print 'processing %s' % encoded_message
            
            if not encoded_message: # sentinel
                if post_handling_func:
                    try:
                        post_handling_func()
                    except:
                        sys.stderr.write('post_handling raised exception.\n')
                        traceback.print_exc()
                        sys.stderr.write('\n')
                        if not sentinel:
                            sentinel = Exception('post_handling errored:\n' + traceback.format_exc())
                
                result_queue.put({'type' : 'sentinel', 'data' : sentinel})
                break
            
            mid = encoded_message['mid']
            cur_retry = encoded_message.get('cur_retry', 0)
            msg_max_retries = encoded_message.get('max_retries')
            if not msg_max_retries:
                msg_max_retries = max_retries
            
            try:
                
                # decodes can crash from import errors
                message, redirect_key = input_queue._decode_message(encoded_message)
                
                try:
                    # construct what we can pass in
                    kwargs = {}
                    if message_handler_argspec.keywords or 'cur_retry' in message_handler_argspec.args:
                        kwargs['cur_retry'] = cur_retry 
                    if message_handler_argspec.keywords or 'max_retries' in message_handler_argspec.args:
                        kwargs['max_retries'] = max_retries
                    
                    results = message_handler_func(message, **kwargs)
                    
                    if not output_queues: # optimization
                        results = None
                    
                    if not expand_iterable_output:
                        # Returning None indicates no message should be output
                        results = [results] if results != None else []
                        
                    # dev note: results may be a user-defined iterator that can raise exceptions
                    for result in results:
                        
                        # each queue uses a unique bucket reference, so must encode multiple times
                        # TODO: This could be optimized to use a single bucket reference for all output queues                    
                        encoded_results = [output_queue._encode_message(result) 
                                           for output_queue in output_queues]                
                        result_queue.put({'type' : 'encoded_result', 'data' : encoded_results})
                        
                
                except Retry as r: # don't intercept with retry_on
                    raise
                except Exception, e: # automatically raise Retries
                    for test_excp in retry_on:
                        if isinstance(e, test_excp):
                            raise Retry(None,None) 
                    
                    raise
                                    
            except Retry as r:
                
                if r.max_retries is not None:
                    msg_max_retries = r.max_retries
                                    
                effective_delay = r.delay
                if effective_delay is None:
                    effective_delay = retry_delay 
                
                if cur_retry == msg_max_retries:

                    # dynamic replace to a MaxRetryException
                    r.__class__ = MaxRetryException

                    print >> sys.stderr, 'Maximum retries reached'
                    traceback.print_exc()
                    print >> sys.stderr, ''
                    if not sentinel:
                        sentinel = Exception('message_handler retries exceeded:\n' + traceback.format_exc())
                    _handle_action_dct_excp(on_error,r,message, cur_retry, msg_max_retries)
                else:                    
                    cloudLog.warning('Retrying on message. current_retries: %d', cur_retry)                    
                    input_queue._push([message], effective_delay, cur_retry+1, msg_max_retries)                
            
            except Exception as e:
                print >> sys.stderr, 'Attached function raised exception. Message ignored.'
                traceback.print_exc()
                print >> sys.stderr, ''
                if not sentinel:
                    sentinel = Exception('message_handler errored:\n' + traceback.format_exc())
                _handle_action_dct_excp(on_error,e,message, cur_retry, msg_max_retries)
                    
                
            # send done message after putting other things on queue
            result_queue.put({'type' : 'ack', 'data' : {'mid' : mid, 'redirect_key' : redirect_key}})       
    
    def attachment():
        """Responsible for creating subprocesses, each of which will have
        run a reader that processes the input queue."""
        
        # Implemented with multiprocessing queue for very high throughput
        # Downside is that we can have one message always buffered
        
        from multiprocessing import Process, Queue as MPQueue
        from multiprocessing.queues import Empty as QueueEmpty
        from multiprocessing.queues import Full as QueueFull
        
        # Beware of setting _update_deadline too close to _deadline
        # Numbers must be set for worst case of queue saturation
        #    (1 long message, then all fast ones after)
        # As acknowledging is ~4x as fast as popping, need:
        #    _update_deadline / (_deadline - _update_deadline) < 4 
         
        _deadline = 60 
        _update_deadline = 30 # processing time when need to update ack 
        
        # inbound queue workers read from
        # max 1 message so main thread waits until retrieving more messages
        message_queue = MPQueue(1) 
        result_queue = MPQueue(1000) # queue where results are written to
        
        input_queue.raise_exceptions = False # in attachment; continue on errors        
        
        all_processes = []
        for _ in xrange(readers_per_job):
            p = Process(target=attachment_runner, args=(message_queue, result_queue))
            p.start()
            all_processes.append(p)
            
        exception = [None] # list of one element so push_to_queue can modify it
        
        processing_tickets = {} # id to ticket_ids, timeout
        finished_tickets = {} # id to ticket_ids
        ticket_lock = threading.Lock()
        ack_cv = threading.Condition()
        process_monitor_cv = threading.Condition()
        
        shutting_down = [False]
                    
        def push_to_queue():
            """Run indefinitely, popping lists of message responses off mp result queue,
            and pushing them to the output cloud queues.
            
            A message response can be a variety of things:
                
                list: Messages to push to queue
                integer: message id: Ack this
                Exception: error sentinel - mark job errored w/ args[0]
                None: Done sentinel - mark job done
                
            """
            
            max_items = 800 # maximum number of items to push to a queue in a single request
            
            sentinel_signals = 0
            
            while sentinel_signals < readers_per_job:
                                
                data_list = [result_queue.get()]
                dequeue_cnt = 1
                
                # flush queue for batching
                try:
                    while dequeue_cnt < max_items:
                        data_list.append(result_queue.get_nowait())
                        dequeue_cnt+=1 
                except QueueEmpty:
                    pass
                
                # extract types
                sentinels = []
                encoded_message_list = []
                ack_dcts = []
                
                type_def = {'sentinel' : sentinels,
                            'encoded_result' : encoded_message_list,
                            'ack' : ack_dcts}
                
                for data_dct in data_list:
                    type_def[data_dct['type']].append(data_dct['data'])
                    
                # sentinels
                sentinel_signals += len(sentinels)
                if not exception[0]:
                    sentinel_errors = [sentinel for sentinel in sentinels if sentinel]
                    if len(sentinel_errors) > 0:
                        exception[0] = sentinel_errors[0]
                                    
                # An encoded_message has 1 to 1 mapping to output queue it is encoded for
                #  If confused what is going on here, ask Aaron
                for output_queue, encoded_message in zip(output_queues, zip(*encoded_message_list)):
                    output_queue._low_level_push( encoded_message )                    

                
                # now that output has been pushed, it is possible to safely acknowledge
                # for performance reasons, the actual ack is done in a different thread
                with ticket_lock:
                    for ack_dct in ack_dcts:
                        ticket_id = ack_dct['mid']                        
                        try:
                            ticket, _ = processing_tickets.pop(ticket_id)
                        except KeyError: # race condition with monitor_pool
                            if not shutting_down[0]:
                                # FIXME: Should initiate shutdown
                                print >> sys.stderr, 'Error: Could not find ticket %s' % ticket_id
                        else:
                            ack_dct['ticket'] = ticket
                            finished_tickets[ticket_id] = ack_dct
                        
                if ack_dcts:
                    with ack_cv: 
                        ack_cv.notify()
                        
                    
        def ack_handler():
            """Submit acknowledgements and deadline updates for processing jobs"""
            while not shutting_down[0] or finished_tickets or processing_tickets:
                
                # when messages rapidly processing, sleep to allow some more acks to come in
                time.sleep(0.5)
                tickets_to_ack = []
                redirect_keys = []
                 
                with ticket_lock:
                    for ticket_dct in finished_tickets.itervalues():
                        tickets_to_ack.append(ticket_dct['ticket'])
                        redirect_key = ticket_dct['redirect_key']
                        if redirect_key:
                            redirect_keys.append(redirect_key)
                    finished_tickets.clear()
                input_queue.ack(tickets_to_ack)
                
                # handle necessary ack updates                
                now = time.time()
                deadline_update_threshold = now + _update_deadline                
                new_timeout = now + _deadline # when ack_handler will wake again
                 
                need_update_tickets = [] 
                with ticket_lock:
                    for mid, (ticket, timeout) in processing_tickets.items()[:]:
                        if timeout < deadline_update_threshold:
                            need_update_tickets.append(ticket)
                            processing_tickets[mid] = (ticket, now + _deadline - 10)
                        else:
                            new_timeout = min(new_timeout,timeout)
                
                if need_update_tickets:
                    input_queue.update_deadline(need_update_tickets, _deadline)
                    
                # Delete bucket references
                # As this supports 1,000 removes per request, it should be fast enough
                try:
                    input_queue._clean_message(redirect_keys)
                except Exception:
                    logging.warning('Could not remove %s' % redirect_keys, exc_info=True)
                        
                
                if not finished_tickets:
                    with ack_cv: # wait until next message needs ack timeout updated
                        ack_cv.wait(min(0.5,time.time() - new_timeout)) 
                            
        def monitor_pool():
            """Monitor queue pool for unexpected terminations.
            As it is difficult to cleanly handle terminations and they are very rare,
                we simply kill the job if this happens.  
                
            FIXME: Very ugly code. This constantly checks if shutting_down[0] == 2 (system shut down);
                aborts once it is
            """        
            while not shutting_down[0]:
    
                with process_monitor_cv:
                    process_monitor_cv.wait(1)
                
                if shutting_down[0]:                        
                    break             
                
                for p in all_processes:
                    if not p.is_alive():
                        wait_time = 22                                                
                        print >> sys.stderr, 'Subprocess unexpectedly terminated. Giving jobs %ds to finish' % wait_time
                        
                        excp = Exception('Subprocess unexpectedly terminated')
                        result_queue.put({'type' : 'sentinel', 'data' : excp},timeout=1)
                        
                        shutting_down[0] = True
                        with process_monitor_cv:
                            # wait for shut down acknowledgement
                            process_monitor_cv.wait(wait_time)
                            
                            # wait for system exit
                            print >> sys.stderr, 'Wait on system exit'
                            process_monitor_cv.wait(wait_time)

                        if shutting_down[0] == 2:
                            print >> sys.stderr, 'shutdown successful'
                            return
                        else:
                            print >> sys.stderr, 'shutdown not yet successful; terminating job'

                        
                        # empty message queue to get main thread moving again 
                        while True:
                            try:
                                message_queue.get(timeout=1)                                
                            except QueueEmpty:
                                break
                        
                        # take down result queue thread
                        try:
                            for _ in xrange(readers_per_job):                                
                                result_queue.put({'type' : 'sentinel', 'data' : excp},timeout=1)
                            
                            # take down acknowledgment handler
                            with ticket_lock:
                                processing_tickets.clear() # "lost" messages will restart elsewhere   
                            
                            cloudLog.info('monitor pool shutting down')
                            break
                        finally: # the BOMG                                                        
                            with process_monitor_cv:
                                process_monitor_cv.wait(wait_time)
                            if shutting_down[0] == 2:
                                return
                            cloudLog.warning('Safe shutdown could not complete. Hard terminating!')
                            for x in all_processes:
                                x.terminate()
                            time.sleep(5.0)
                            
                            if shutting_down[0] == 2:
                                return                            
                            thread.interrupt_main()
                            time.sleep(2.0)
                            
                            if shutting_down[0] == 2:
                                return                            
                            cloudLog.warning('Monitor aborting process!')
                            os._exit(1)  
                            cloudLog.warning('This line should not have run..')
                        
                    
        push_thread = threading.Thread(target=push_to_queue)
        push_thread.daemon = True
        push_thread.name = 'Output Queue Push Thread'
        push_thread.start()

        ack_thread = threading.Thread(target=ack_handler)
        ack_thread.daemon = True
        ack_thread.name = 'Ack Handler Thread'
        ack_thread.start()

        monitor_thread = threading.Thread(target=monitor_pool)
        monitor_thread.daemon = True
        monitor_thread.name = 'Monitor pool Thread'
        monitor_thread.start()

        cloudLog.debug('All subprocesses started. Listening on queue')
        
        # Main thread processes the input queue
        message_id = 1
        while not shutting_down[0]:
            messages = input_queue._low_level_pop_tickets(batch_size, timeout=_job_shutdown_timeout,
                                                           deadline=_deadline)
            #print 'got messages %s' % messages
            if not messages:  # shut down job
                break 
            
            # inject unique identifiers into message            
            with ticket_lock:
                now = time.time()
                for message in messages:
                    message['mid'] = message_id                    
                    processing_tickets[message_id] = (message['ticket'], now + _deadline - 10)
                    message_id += 1                    
            
            try:
                for message in messages:
                    del message['ticket']
                    message_queue.put(message)
            except QueueFull, qe: # on system exit, this may be raised; translate to a system exit
                cloudLog.info('Got QueueFull (%s) exception. Possibly kill signal?' % str(qe))
                raise SystemExit('Killed')
        
        # SHUTDOWN
        req_shutdown = shutting_down[0]
        if req_shutdown:
            cloudLog.debug('shutdown requested')
        else:
            cloudLog.debug('No more messages in queue -- shutting down')                    
        
        shutting_down[0] = True
        with process_monitor_cv:
            process_monitor_cv.notify()        
        
        # send sentinels
        sentinels_to_send = len(all_processes)
        while sentinels_to_send:
            try:
                message_queue.put(None, timeout=5)
            except QueueFull:
                alive_processes = [p for p in all_processes if p.is_alive()]
                if not alive_processes:
                    break
            else:
                sentinels_to_send-=1

        cloudLog.debug('Waiting for subprocesses to finish')

        # wait for all processes to get sentinel and terminate
        while [p for p in all_processes if p.is_alive()]:
            for p in all_processes:
                p.join(10)
        
        cloudLog.debug('Waiting for outputs to flush.')                    
        push_thread.join()
        
        cloudLog.debug('Waiting for validation to flush.')
        with ack_cv:
            ack_cv.notify() # force wake up
        ack_thread.join()
        
        cloudLog.debug('Done')
        
        shutting_down[0] = 2 # successful shut down
        
        with process_monitor_cv:
            process_monitor_cv.notify()        
        
        if exception[0]:
            raise CloudException(exception[0])
        
    attachment()



"""
Science VM configuration backend

Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2010 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see
http://www.gnu.org/licenses/lgpl-2.1.html
"""

import os
import sys

class NoOptionError(Exception):
    """A requested option was not found."""

    def __init__(self, option):
        Exception.__init__(self, "No key %r" % option)
        self.option = option

extraInfo = {
             'Account': 'PiCloud account information. This is the only section that you need to worry about.',
             'Logging': 'Control what should be logged and where',
             'Transport': 'PiCloud information transfer',
             'Multiprocessing': 'Options that control running the cloud locally',
             'Simulation': 'Options for simulation mode that override Multiprocessing and Logging options'
        }


class ConfigManager(object):

    backend = None
    hiddenSets = []

    @staticmethod
    def getCommentStr(section, option):
        return option.lower()

    def __init__(self, defaults=None):
        self.sections = {}
        self.optioncomment = {}

    def read(self, fname):
        """Return True on successful read"""
        import os
        import sys
        dir = os.path.dirname(fname)
        conf = os.path.basename(fname)
        pyfile = os.path.splitext(conf)[0]
        addedEntry = False

        try:
            if dir not in sys.path:
                sys.path.append(dir)
                addedEntry = True
            if not os.path.exists(fname):
                try:
                    os.unlink("".join([dir, os.sep, pyfile, '.pyc'])) #force recompilation
                except OSError:
                    pass
                import types
                self.backend = types.ModuleType('cloudconf')
                return False #force rewrite
            else:
                try:
                    if pyfile in sys.modules:
                        self.backend = sys.modules[pyfile]
                    else:
                        self.backend = __import__(pyfile)
                except ImportError, e:
                    import types
                    sys.stderr.write('CLOUD ERROR: Malformed cloudconf.py:\n %s\nUsing default settings.\n' % str(e))
                    self.backend = types.ModuleType('cloudconf')
        finally:
            if addedEntry:
                sys.path.remove(dir)
        return True

    def get(self, section, option, comment = None):
        if not hasattr(self.backend, option):
            raise NoOptionError(option)
        value = getattr(self.backend, option)
        self.sections.setdefault(section, {})[option] = value
        if comment:
            self.optioncomment[self.getCommentStr(section, option)] = comment
        return value

    def hiddenset(self, *args):
        """Defer set commands"""
        self.hiddenSets.append(args)

    def showHidden(self):
        """Do all deferred (hidden) sets -- not thread safe"""
        for hiddenSet in self.hiddenSets:
            self.set(*hiddenSet)
        self.hiddenSets = []

    def set(self, section, option, value, comment = None):
        self.sections.setdefault(section, {})[option] = value
        if comment:
            self.optioncomment[self.getCommentStr(section, option)] = comment
        #print 'setting backend %s to %s' % (option, value)
        setattr(self.backend,option,value)

    def write(self, fp):
        """Write configuration file with defaults
        Include any comments"""
        #hack to ensure account comes first:
        sections = self.sections.keys()
        sections.sort()

        for section in sections:
            cmt = '"' * 3
            fp.write('%s\n%s\n' % (cmt, section))
            ei = extraInfo.get(section)
            if ei:
                fp.write('%s\n%s\n' % (ei, cmt))
            else:
                fp.write('%s\n' % cmt)
            started = False
            for (key, value) in self.sections[section].items():
                if key != "__name__":
                    comment = self.optioncomment.get(self.getCommentStr(section, key))
                    if comment:
                        if started:
                            fp.write('\n')
                        for cel in comment.split('\n'):
                            fp.write('# %s\n' % cel.strip())
                    #print 'write %s=%s with type %s'% (key, repr(value), type(value))
                    fp.write("%s = %s\n" %
                             (key, repr(value).replace('\n', '\n\t')))
                    started = True
            fp.write("\n\n")

class ConfigSettings(object):
    """This object provides the ability to programmatically edit the cloud configuration (found in cloudconf.py).
    ``commit()`` must be called to update the cloud module with new settings - and restart all active clouds
    """


    @staticmethod
    def _loader(path,prefix, do_reload):
        """Bind """
        files = os.listdir(path)
        delayed = []
        for f in files:
            if f.endswith('.py'):
                endname = f[:-3]
                if endname == 'cloudconfig' or endname == 'configmanager' or endname == 'setup' or endname == 'writeconfig' or endname == 'cli':
                    continue
                if endname == '__init__':
                    delayed.append(prefix[:-1])  #do not load __init__ until submodules reloaded
                    continue
                elif endname == 'mp':
                    modname = prefix + endname
                    delayed.append(modname)
                else:
                    modname = prefix + endname
                #print modname #LOG ME
                if do_reload:
                    if modname in sys.modules:
                        try:
                            reload(sys.modules[modname])
                        except ImportError:
                            pass
                else:
                    try:
                        __import__(modname)
                    except ImportError:
                        pass
            elif os.path.isdir(path + f):
                newpath = path + f + os.sep
                ConfigSettings._loader(newpath,prefix + f + '.',do_reload)
        if delayed:
            if '__init__' in delayed: #must come last
                delayed.remove('__init__')
                delayed.append('__init__')

            for delay_mod in delayed:
                if do_reload:
                    if delay_mod in sys.modules:
                        try:
                            reload(sys.modules[delay_mod])
                        except ImportError:
                            pass
                else:
                    try:
                        __import__(delay_mod)
                    except ImportError:
                        pass
            delayed = []



    def _showhidden(self):
        """Show hidden variables"""
        self.__confmanager.showHidden()
        self.__init__(self.__confmanager) #restart


    def commit(self):
        """Update cloud with new settings.

         .. warning::

            This will restart any active cloud instances, wiping mp/simulated jobs and setkey information
        """
        import scicloud as cloud
        setattr(cloud,'__immutable', False)
        cloud.cloudinterface._setcloud(cloud, type=None)
        if hasattr(cloud,'mp'):
            setattr(cloud.mp,'__immutable', False)
            cloud.cloudinterface._setcloud(cloud.mp, type=None)

        #Reload cloud modules in correct order
        mods = cloud._modHook.mods[:]

        for modstr in mods:
            mod = sys.modules.get(modstr)
            if mod and modstr not in ['cloud.util.configmanager', 'cloud.cloudconfig']:
                try:
                    reload(mod)
                except ImportError:
                    pass
        reload(cloud)
        cloud._modHook.mods = mods #restore mods after it is wiped

    def __init__(self, confmanager, do_reload=False):
        backend = confmanager.backend
        self.__confmanager = confmanager

        def _set_prop(item):
            if hasattr(backend, item):
                typ = type(getattr(backend, option))
                if typ is type(None):
                    typ = None
            else:
                typ = None
            #print 'item %s has type %s' % (item, typ)
            def __inner__(self, value):
                if typ:
                    try:
                        k = typ(value)
                        setattr(backend,item, k)
                    except ValueError, e:
                        raise ValueError('Configuration option %s must have type %s.' % (option, typ.__name__))

            return __inner__

        def _get_prop(item):
            def __inner__(self):
                return getattr(backend, item)
            return __inner__

        import scicloud as cloud
        ConfigSettings._loader(cloud.__path__[0] + os.sep ,'cloud.',do_reload)
        for options in confmanager.sections.values():
            for option in options:
                prop = property(_get_prop(option), _set_prop(option), None, confmanager.optioncomment.get(ConfigManager.getCommentStr("",option)))
                setattr(self.__class__, option, prop)





"""
Binds cloud object methods to outer modules

Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

from . import cloudconfig as cc

#globals for cloud.files
last_api_key = None
last_api_secretkey = None


def generate_cloud(type='network', api_key=None, api_secretkey=None, server_url=None):
    """
    Generate a new Cloud object.  Generally setCloud should be used;
    This function should only be used if you wish to have duplicate clouds 
    running
    
    type is a string that specifies which cloud to use.
    The two types are
    type='network' (The cloud service provided by PiCloud. *Default*)
    type='local' (Simulated mode - Runs functions locally)
    
    If type is set to 'network', api_key and api_secretkey must be
    specified.  See your PiCloud account API Keys for this information
    """
    from .cloud import Cloud
    
    if type == 'network':
        from .transport import HttpConnection
        from .transport import DependencyAdapter as CloudAdapter                                
        retval =  Cloud(CloudAdapter(HttpConnection(api_key, api_secretkey, server_url=server_url)))
        global last_api_key, last_api_secretkey
        last_api_key = api_key
        last_api_secretkey = api_secretkey
        return retval
    elif type == 'simulated':   
        try:
            from .transport.simulator import SimulatedConnection
        except ImportError:
            try:
                import multiprocessing
            except ImportError: #if mp error, fall back
                from .cloudlog import cloudLog
                cloudLog.error('Multiprocessing is not installed. Simulator cannot be started. Falling back to regular PiCloud')
                return generate_cloud()
            else:
                raise #something else has gone wrong    
        from .transport import SerializingAdapter                          
        return Cloud(SerializingAdapter(SimulatedConnection()))
    elif type == 'mp': #remove me
        from .transport import SerializingAdapter                                        
        #we check for mp existance in __init__ to allow module to be removed
        from .transport.local import MPConnection
        return Cloud(SerializingAdapter(MPConnection()))
    else:
        raise Exception('cloud: Cloud generation type must be either network or local, not %s' % type)                            


def _setcloud(module, type='network', api_key=None, api_secretkey=None, server_url=None, restart=False, immutable=False):
    """
    Set the cloud that cloud should use
    Binds module's cloud to this cloud
    
    type is a string that specifies which cloud to use.
    The two types are:
    type='network' (Default: The cloud service provided by PiCloud.)
    type='local' (Simulated mode - Runs functions locally)
    
    If type is set to 'network', api_key and api_secretkey must be
    specified.  See your PiCloud account API Keys for this information    
    
    restart is used internally to force the cloud to be reconstructed and 
    reconnected
    
    If immutable is set, future setCloud calls will have no effect        
    """
    
    if not module.__immutable and (type != module.__type or restart or not module.__cloud or 
       module.__cloud.needs_restart(api_key=api_key,api_secretkey=api_secretkey, server_url=server_url)):
        
        # close any existing cloud
        if module.__cloud != None and module.__cloud.opened:
            module.__cloud.close()
                
        if type:
            cloud = generate_cloud(type, api_key, api_secretkey, server_url)        
            _bindcloud(module,cloud,type, immutable)
        else:
            module.__cloud = None
            

def _bindcloud(module, cloud, type, immutable=False):
    module.__cloud = cloud 
    
    module.__type = type
    
    #bind every cloud method listed in module's __all__ to module 
    for meth in module.__all__:
        if hasattr(cloud,meth):
            setattr(module,meth,getattr(cloud,meth))
    
    cloud.parentModule = module.__name__
    
    if immutable:
        module.__immutable = True 
    
    
    



"""
A debugging pickle object
This provides an xml trace of a pickle

Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

import os
import types
import sys

import functools
import pickle
from pickle import PicklingError
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from . import xmlhandlers
from .cloudpickle import CloudPickler
from .. import util
from ..util import islambda
from .. import cloudconfig as cc

import logging
cloudLog = logging.getLogger("Cloud.Transport")

class DebugPicklingError(pickle.PicklingError):
    def __init__(self, msg, xmltrace):        
        pickle.PicklingError.__init__(self,msg)
        self.xmltrace = xmltrace
        
    def __str__(self):        
        return 'Pickle trace:\n' + self.xmltrace + '\n' + self.args[0] 


class CloudDebugPickler(CloudPickler):    
    
    typeAttributeFunctions = {} #type based attribute function
    typeCustomChildPrinter = {} #custom type printers for children
    
    #other properties:
    
    showPrimitives = cc.logging_configurable('show_primitives',
                                                       default=False,
                                                       comment="Should primitives be logged?",hidden=True)     
    printingMinSize = cc.logging_configurable('printing_min_size',
                                                        default=0,
                                                        comment="Minimum size objects should be to be logged",hidden=True)    
    showTransmittedTypesComposition =  \
        cc.logging_configurable('show_transmitted_types_composition',
                                       default=False,
                                       comment="Should composition of transmitted types be logged?",hidden=True)
        
    transmitted_forced_imports = None
    
    min_size_to_save = 0 #do not hold objects < min_size_to_save  
      
    
    def __init__(self, file, protocol=None):        
        CloudPickler.__init__(self,file,protocol)
        self.file = file
        self.traceStack = [] #stack to track as objects are added
        self.objSizes = {} #sizes of objects after pickling including sub data 
        self.objChildren = {} #objects pickled under this ID
        self.lastobj = None   
        self.aborting = False                            
    
    def dump(self, obj):
        
        try:
            CloudPickler.dump(self,obj)
        except PicklingError, p:            
            cloudLog.exception("pickling failed on %s of type %s." 
                           % (self.lastobj, type(self.lastobj)))
            strio = StringIO()                           
            
            xmlf = xmlhandlers.XmlWriter(strio,header=False)
            xmlmid = xmlhandlers.XmlStackWriter(xmlf)
            #xmlmid = xmlf
            
            self.aborting = True
            try: #intentionally crash at partial xml            
                self.dump_obj(xmlmid, obj, self.print_top_level, None, {}, abortObject = self.lastobj)                        
            except DebugPicklingError, e:
                if e.args[0] != 'fake exit':
                    raise     
            except RuntimeError, e: # dump_obj will crash w/ infinite recursion if regular dump did
                if 'recursion' not in e.args[0]:
                    raise
                                       
            self.aborting = False                
            xmlmid.flush()
            #raise
            raise DebugPicklingError(p.args[0],strio.getvalue())
        else:
            self.lastobj = None            
        
    
    def save_function_tuple(self, func, forced_imports):
        CloudPickler.save_function_tuple(self, func, forced_imports)
        if forced_imports:
            self.transmitted_forced_imports = forced_imports
    
    def save(self, obj):        
        #ignored types:        
        t = type(obj)
        
        #print 'saving...', len(self.traceStack), obj, type(obj)
        
        if id(obj) in self.memo: #deal with this better!!!            
            CloudPickler.save(self,obj)
            return
        
        self.lastobj = obj
        #block Numeric and boolean objects
        self.traceStack.append(id(obj))
        startpos = self.file.tell()
        
        #Do save:
        excp = None        
        try:            
            CloudPickler.save(self,obj)
        except PicklingError, p:               
            excp = p
            if id(obj) not in self.memo:  #forcibly memoize objects for debug printing..
                self.memoize(obj)

        endpos = self.file.tell()
        self.traceStack.pop()
        
        size = endpos - startpos
        ido = id(obj)
        if excp or size >= self.min_size_to_save:  #save if exception (to print debug) or larger than min_size
            self.objSizes[ido] = size
            if self.traceStack:        
                idp = self.traceStack[-1]
                if idp not in self.objChildren:
                    self.objChildren[idp] = []
                self.objChildren[idp].append(ido)
            if excp:
                raise excp
    
    
    """
    Reporting...
    Returns if an element was pushed
    """
    
    """XML Element Printers"""
    def print_dependency(self, xmlf, printerdata, obj, attrs):
        is_memo = attrs.get('is_memoized')         
        if not is_memo:  #hide memoized deps          
            attrs['objId'] = str(id(obj))              
            xmlf.push("Dependency",attrs)
            return True
        return False
        
    
    def print_class_attribute(self, xmlf, printerdata, obj, attrs):
        attrs.insertAt(0,'name',printerdata)
        xmlf.push('Attribute',attrs)
        return True

    def print_nontransmitted_attribute(self, xmlf, printerdata, obj, attrs):
        attrs.insertAt(0,'name',printerdata)
        xmlf.push('NonTransmittedAttribute',attrs)
        return True

    
    def print_collection(self, xmlf, printerdata, obj, attrs):
        """tuples and lists"""
        attrs.insertAt(0,'entryNum',str(printerdata))
        xmlf.push('Element',attrs)
        return True
    
    def print_dictionary_element(self, xmlf, printerdata, obj, attrs):
        """Dictionary"""
        attrs.insertAt(0,'key',printerdata[0])
        attrs.insertAt(1,'keyObjId',str(printerdata[1]))
        xmlf.push('Value',attrs)
        return True
                       

    def print_top_level(self, xmlf, printerdata, obj, attrs):
        xmlf.push("PickledObject",attrs)
        return True

    
    """
    Below methods define handlers to set control xml attributes for different types
    Each handler receives an the object to dump and XML element attributes to minipulate
    It returns childAttributes, allowDeps, childPrinter
    ChildAttributes is a dictionary of key/value pairs. the key is passed into the childPrinter and the value
        is the actual child
    If allowDeps is true, any dependencies this object has (outside of the childAttributes) are printed
    childPrinter is one of the before-defined xml printers that should be run
    
    childAttributes, allowDeps, childPrinter = attrFunction(self,obj, attrs)
    """
    
    def write_function(self, obj, attrs):        
        #attrs['type'] = "functionReference"         
        
        clist = self.objChildren.get(id(obj))
        
        if islambda(obj):            
            attrs['isLambda'] = str(True)
        else:
            attrs['funcName'] = str(obj.func_code.co_name)
        attrs['module'] = str(obj.__module__)            
        attrs['fileName'] = obj.func_code.co_filename
        attrs['firstLineNumber'] = str(obj.func_code.co_firstlineno)  
        if obj.func_code.co_freevars:
            attrs['closureVariables'] = pretty_print_str_iterator(obj.func_code.co_freevars) 

        #serialized function information:
        if clist:
            if len(clist) == 2:  #save global with attribute likely               
                m = self.memo.get(clist[1])[1]
                if isinstance(m, dict):
                    return m, False, self.print_class_attribute
            elif len(clist) > 2:                 
                attrs['TransmittedFunctionCode'] = str(True)
                #search for attributes. Expected ordering
                #code, globals, defaults closure, dict.  all but code can be missing
                #CodeType, dict, list, tuple/None, dict expected 
                i = 0
                clist = map(lambda x: self.memo.get(x,(None,None))[1],clist)
                max = len(clist)
                codeobj = None
                while i < max:                    
                    tst = clist[i]                                   
                    if type(tst) is types.TupleType and len(tst) == 2 and type(tst[0]) is types.CodeType:
                        codeobj = tst[1]
                        i+=1
                        break
                    i+=1
                if i == max: #extremely rare if even possible                    
                    return {}, True, self.print_class_attribute
                outdict = {}
                outdict['code'] = codeobj
                
                found = 0
                while i < max:
                    tst = clist[i]
                    typtst = type(tst)
                    if found <= 0 and typtst is types.DictType: #globals?
                        if i == max -1 and not obj.func_code.co_names: #might be attributes??
                            return {}, True, self.print_class_attribute #just list dependencies
                        #could still be attributes, but highly unlikely
                        outdict['globals'] = tst
                        found = 1
                    elif found <= 1 and typtst is types.TupleType: #defaults!
                        outdict['defaults'] = tst
                        found = 2
                    elif found <= 2 and typtst is types.ListType: #closure!
                        outdict['closure'] = tst
                        found = 3
                    elif found <= 3 and typtst is types.DictType:
                        outdict['function_attributes'] = tst
                        found = 4
                        break
                    i+=1
                
                return outdict, False, self.print_class_attribute
            
        return {}, True, self.print_class_attribute

    typeAttributeFunctions[types.FunctionType] = write_function
        
    def write_method(self, obj, attrs):
        """A method is a function bound to a class"""
        return {'function':obj.im_func,
                'self':obj.im_self,
                'class': obj.im_class},True,self.print_class_attribute
    typeAttributeFunctions[types.MethodType] = write_method            
       

    def write_partial(self, obj, attrs):
        return {'function': obj.func,
                'args': obj.args,
                'kwargs': obj.keywords
                },False,self.print_class_attribute

    typeAttributeFunctions[functools.partial] = write_partial

       
    def write_file(self, obj, attrs):
        """Transmitted file object"""
        attrs['filename'] = os.path.abspath(obj.name)
        return {},False,None
    typeAttributeFunctions[types.FileType] = write_file            
    
    
    def write_type_descriptor(self, obj, attrs):
        clist = self.objChildren.get(id(obj))
        if clist:
            attrs['transmittedDescriptor'] = str(True)   
            
        repr = str(obj)
        if repr[0:5] == '<type':
            attrs['type'] = 'typeDescriptor'
            attrs['typeDescribed'] = str(obj)[7:-2]
            #attrs['module'] = obj.__module__
            return {}, self.showTransmittedTypesComposition or self.aborting, None #Block out info?
        elif repr[0:6] == '<class':
            attrs['type'] = 'classDescriptorNew'
            attrs['classDescribed'] = str(obj)[8:-2]
            #attrs['module'] = obj.__module__
            return {}, self.showTransmittedTypesComposition or self.aborting, None #I'm not sure if this is how to do this?
        else: #???            
            #print 'wtf??', repr
            return {}, True, None    
    typeAttributeFunctions[types.TypeType] = write_type_descriptor
    
    def write_class_descriptor(self, obj, attrs):
        attrs['type'] = 'classDescriptorOld'
        attrs['classDesribed'] = str(obj)    
        #return obj.__dict__, True, self.print_class_attribute
        
        #check for type transmision:
        clist = self.objChildren.get(id(obj))
        if clist:
            attrs['transmittedDescriptor'] = str(True)        
        return {}, self.showTransmittedTypesComposition or self.aborting, None    
    typeAttributeFunctions[types.ClassType] = write_class_descriptor
    
    def write_raw_code(self, obj, attrs):
        #Suppress code blocks
        #attrs['type'] = 'pythonByteCode'
        return {}, False, None
    typeAttributeFunctions[types.CodeType] = write_raw_code

    def write_string(self, obj, attrs):
        if not self.showPrimitives and len(obj) < 256:
            attrs['hide'] = True
        if len(obj) < 1024:        
            attrs['contents'] = obj
        else:
            attrs['contents'] = obj[:1021] + '...'
        return {}, False, None
    typeAttributeFunctions[str] = write_string
    typeAttributeFunctions[unicode] = write_string
    
    def write_numeric(self, obj, attrs):    
        if attrs.has_key('implicitly_pickled'): #int's are generally not memoized - don't report it though
            del attrs['implicitly_pickled']
        if not self.showPrimitives:
            attrs['hide'] = True                
        attrs['contents'] = str(obj)                    
        return {}, False, None
    typeAttributeFunctions[int] = write_numeric
    typeAttributeFunctions[float] = write_numeric
    
    def write_complex(self, obj, attrs):    
        if not self.showPrimitives:
            attrs['hide'] = True                
        attrs['contents'] = str(obj.real) + '+' + str(obj.imag) + 'j'                    
        return {}, False, None    
    typeAttributeFunctions[complex] = write_complex

    
    def write_module(self, obj, attrs):
        attrs['module'] = obj.__name__
        return {}, False, None
    typeAttributeFunctions[types.ModuleType] = write_module
    
    def get_type_name(self, typ):
        if typ.__module__ != '__builtin__':
            return typ.__module__ + '.' + typ.__name__
        else:
            return typ.__name__
    
    def determine_collection_type(self, obj):
        """Determine the type of a sequence"""
        """Returns TYPE or MIXED"""        
        if not obj:
            return 'Empty'
        ctype = False
        for o in obj:
            if ctype is False:
                ctype = type(o)                
            elif type(o) != ctype:
                return 'Mixed'
        return self.get_type_name(ctype)
    
    def attr_collection(self, obj, attrs):
        attrs['numElements'] = str(len(obj)) 
        if obj:
            attrs['containedType'] = self.determine_collection_type(obj)
        i = 0
        outAttr = {}
        for c in obj:
            outAttr[i] = c
            i+=1
        return outAttr

    
    def write_sequence(self, obj, attrs):
        return self.attr_collection(obj,attrs),True,self.print_collection    
    typeAttributeFunctions[tuple] = write_sequence
    typeAttributeFunctions[list] = write_sequence


    def write_set(self, obj, attrs):
        return self.attr_collection(obj,attrs),False,self.print_collection   
    typeAttributeFunctions[set] = write_set
    typeAttributeFunctions[frozenset] = write_set
    
    def write_dictionary(self, obj, attrs):
        attrs['numElements'] = str(len(obj)) 
        if obj:
            attrs['keyType'] = self.determine_collection_type(obj.keys())
            attrs['valueType'] = self.determine_collection_type(obj.values())
        outAttr = {}
        for key, value in obj.items():
            typ = type(key)
            if typ == int or typ==float:
                attrname = str(key), id(key)
            elif typ == str:
                if len(key) > 128:
                    attrname = key[:64] + '...' + key[-64:], id(key)
                else:
                    attrname = key, id(key)
            else:
                attrname = 'object', id(key)
            outAttr[attrname] = value            
        return outAttr,True,self.print_dictionary_element    
    typeAttributeFunctions[dict] = write_dictionary
    
    def write_image(self, obj, attrs):
        """PIL Images"""
        attrs['type'] = 'PIL Image'
        attrs['resolution'] = str(obj.size)
        attrs['mode'] = str(obj.mode)
        if not obj.im and obj.fp and 'r' in obj.fp.mode and obj.fp.name \
            and not obj.fp.closed and (not hasattr(obj, 'isatty') or not obj.isatty()):
            attrs['transmitted_raw_data'] = str(False)
            return {'file': obj.fp}, False, self.print_class_attribute
        else:
            attrs['transmitted_raw_data'] = str(True)
            return {}, False, None
            
        
    
    def write_instance(self, obj, attrs):
        """For an old style class"""
        
        #hook for images:
        if hasattr(obj,'im') and hasattr(obj,'palette') and 'Image' in obj.__module__:
            return self.write_image(obj,attrs)
        
        #just handle state:
        try:
            getstate = obj.__getstate__
        except AttributeError:
            stuff = obj.__dict__
        else:
            stuff = getstate()
            
        if type(stuff) is not dict:
            stuff = {}
        return stuff, True, self.print_class_attribute
    typeAttributeFunctions[types.InstanceType] = write_instance
            
                
    def write_user_class(self, obj, attrs):
        """For a user defined new-style class - default"""
        if not issubclass(type(obj), types.TypeType) and hasattr(obj, '__module__') and hasattr(obj, '__reduce_ex__'):
            """New-style user class"""
            redf= obj.__reduce_ex__(2)
            if len(redf) >= 2 and getattr(redf[0],'__name__',"") == "__newobj__":
                if len(redf) >= 3:
                    attr = redf[2]
                if type(attr) is not dict:
                    attr = {}
                return attr, True, self.print_class_attribute
        #default behavior..
        if hasattr(obj,'__module__'):
            attrs['module'] = str(obj.__module__)            
        return {}, True, None
        
    def dump_obj(self, xmlf, obj, printer, printerdata, localMemo, abortObject = None):
        """Print an object
        xmlf is an XmlWriter
        obj is the object to print
        printer is a method describing how to print this object
        printerdata is data used by the printer object
        if memoized_ref is true, a reference is printed
        """       
        #print obj, abortObject, abortObject is obj
        typ = type(obj)
        attrs = util.OrderedDict()
        ido = id(obj)                    
        sz = self.objSizes.get(ido)
        if sz == None: #todo... return?
            #print 'null sz'
            sz = 0 
        if sz < self.printingMinSize and (not abortObject or abortObject is not obj): #hide small objects
            return

        memo_entry = self.memo.get(ido)        
        
        attrs['type'] = self.get_type_name(type(obj))
        
        memoized_ref = ido in localMemo
        if sz == 0: #non-physical object
            attrs['not_pickled'] = str(True)
        elif obj is None:
            pass
        elif not memo_entry:  #implicit pickling -- reduced to another object
            attrs['implicitly_pickled'] = str(True)
            if memoized_ref:
                attrs['is_memoized'] = str(True)
            else:
                attrs['size'] =  str(sz) 
                localMemo[ido] = True        
        elif memoized_ref:
            attrs['is_memoized'] = str(True)
            attrs['memo_id'] = str(localMemo[ido])  #memo_len identiy                
        else:      #new object            
            attrs['size'] =  str(sz)     
            if memo_entry is not None:
                attrs['memo_id'] = str(memo_entry[0])  #memo_len identiy
                localMemo[ido] = memo_entry[0]            
                                            
        attrFunction = self.typeAttributeFunctions.get(typ)   
        #print 'typ', typ, attrFunction     
        if attrFunction:
            childAttributes, allowDeps, childPrinter = attrFunction(self,obj, attrs)
        else:                       
            childAttributes, allowDeps, childPrinter = self.write_user_class(obj,attrs)
            #print 'user', childAttributes, allowDeps, childPrinter      
        
        """
        if isinstance(childAttributes, tuple):
                    print childAttributes, attrFunction.__name__, typ
        """ 
        
        if abortObject and abortObject is obj:
            attrs.insertAt(0,'failed_to_pickle',str(True))            
        elif attrs.get('hide'):
            return
             
        needPop = printer(xmlf, printerdata, obj, attrs)  #push xml element
        
        if abortObject and abortObject is obj:
            #print 'AB',  abortObject
            raise DebugPicklingError('fake exit',None)
        
        if not memoized_ref:
            customPrinter = self.typeCustomChildPrinter.get(typ)
            if customPrinter is not None: 
                #custom Printers are responsible for all child serialization
                customPrinter(self,xmlf,obj)
            else:                          
                #function below??:            
                clearedDeps = [] #list of dependency ids already printed as attributes        
                clist = self.objChildren.get(ido) #list of dependencies of this object 
                #print '\nclist', clist
                #attributes:      
                for key, value in childAttributes.items():
                    idc = id(value)
                    self.dump_obj(xmlf,value,childPrinter,key,localMemo,abortObject)
                    clearedDeps.append(idc)
                if allowDeps:  #other dependencies..
                    if clist:
                        for idc in clist:
                            if idc in clearedDeps:
                                continue                            
                            child_memo_entry = self.memo.get(idc)                            
                            if child_memo_entry is not None:
                                cobj = child_memo_entry[1]
                                #validate tuples/lists/dicts:
                                if type(cobj) == tuple or type(cobj) == list:
                                    if all([id(subobj) in clearedDeps for subobj in cobj]):
                                        continue
                                elif type(cobj) == dict:
                                    #verify if this is an attribute list
                                    if all([type(key) == str for key in cobj.keys()]):
                                        if all([id(subobj) in clearedDeps for subobj in cobj.values()]):
                                            continue                                    
                                self.dump_obj(xmlf,cobj,self.print_dependency,None,localMemo,abortObject)
        if needPop:
            if printer == self.print_top_level:
                #always show transmitted_forced_imports at top level
                if self.transmitted_forced_imports:
                    xmlf.push("forced_imports")
                    for mod in self.transmitted_forced_imports:
                        xmlf.push('item',{'module':mod.__name__})
                        xmlf.pop()
            xmlf.pop() #pop printed entry
    
    def write_report(self, topobj, outfile, abortObject = None, hideHeader=False):
        """Write an xml report about the pickling of topobj to outfile"""           
        if not abortObject:  #if there is no explicit abortObject use lastobj (if any)
            abortObject = self.lastobj                 
        xmlf = xmlhandlers.XmlWriter(outfile,header=not hideHeader)
        if not hideHeader:
            xmlf.comment('Properties',{'showPrimitives':str(self.showPrimitives),
                                   'minimumObjectSizePrinted':str(self.printingMinSize),
                                   'typeCompositionShown':str(self.showTransmittedTypesComposition)})                                   
        self.aborting = True
        try:
            self.dump_obj(xmlf, topobj, self.print_top_level, None, {}, abortObject = abortObject)
        except DebugPicklingError: #ignore aborts
            pass
        self.aborting = False
        
def pretty_print_str_iterator(str_iterator):
    k = '['
    start = False
    for str in str_iterator:
        if start:
            k+=','
        k+=str
        start = True
        
    k+=']'
    return k

#shorthands for legacy support

def dump(obj, file, protocol=2):
    CloudDebugPickler(file, protocol).dump(obj)

def dumps(obj, protocol=2, debugprint = False):    
    file = StringIO()        
    cp = CloudDebugPickler(file, protocol)
    cp.dump(obj)
        
    #if type(obj) is not dict and type(obj) is not list and type(obj) is not tuple:
    if debugprint:     
        print 'reporting', type(obj), obj
        cp.write_report(obj,sys.stdout) 
    return file.getvalue()




"""
This module holds convenience functions for accessing information

Copyright (c) 2014 `Science Automation Inc. <http://www.scivm.com>`_.  All rights reserved.

email: support@scivm.com

Copyright (c) 2013 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@picloud.com

The cloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see
http://www.gnu.org/licenses/lgpl-2.1.html
"""

import datetime
import logging
import os
import platform
import re
import sys
import time

from subprocess import Popen, PIPE, STDOUT

try:
    import json
except:
    # Python 2.5 compatibility
    import simplejson as json

import scicloud as cloud
from . import credentials
from ..cloudlog import stdout as print_stdout

cloudLog = logging.getLogger('util.common')


########## Utilities for sending post request ############
def _send_request(request_url, data, jsonize_values=True):
    """Makes a cloud request and returns the results.

    * request_url: whee the request should be sent
    * data: dictionary of post values relevant to the request
    * jsonize_values: if True (default), then the values of the *data*
        dictionary are jsonized before request is made."""
    if jsonize_values:
        data = _jsonize_values(data)
    conn = cloud._getcloudnetconnection()
    return conn.send_request(request_url, data)

def _jsonize_values(dct):
    """Jsonizes the values of the dictionary, but not the keys."""
    jsonized = [(key, json.dumps(value)) for key, value in dct.items()]
    jsonized.append(('_jsonized', True))
    return dict(jsonized)

def _fix_time_element(dct, keys):
    if not hasattr(keys, '__iter__'):
        keys = [keys]

    FORMAT = '%Y-%m-%d %H:%M:%S'
    for key in keys:
        if dct.has_key(key):
            val = dct.get(key)
            dct[key] = (None if (val == 'None')
                        else datetime.datetime.strptime(val, FORMAT))

    return dct


########## Platform dependencies ###########
plat = platform.system()

def _rsync_path():
    base_dir = os.path.join(sys.prefix, 'extras') if plat == 'Windows' else ''
    return os.path.join(base_dir, 'rsync')

def _ssh_path():
    base_dir = os.path.join(sys.prefix, 'extras') if plat == 'Windows' else ''
    return os.path.join(base_dir, 'ssh')

def _check_rsync_dependencies():
    """Checks dependencies by trying the commands."""
    msg_template = 'Dependency %s not found'
    if plat == 'Windows':
        msg_template += ' in cloud installation'

    # check rsync exists
    status, _, _ = exec_command('%s --version' % _rsync_path(), pipe_output=True)
    if status:
        raise cloud.CloudException(msg_template % 'rsync')

    # check ssh exists
    status, _, _ = exec_command('%s -V' % _ssh_path(), pipe_output=True)
    if status:
        raise cloud.CloudException(msg_template % 'ssh')

class WindowsPath(object):
    """Processes Windows paths for easy conversion into cwRsync compatible
    format.  This involves replacing drive specification into
    /cygdrive/[drive_lettter] format, and replacing backward slashes with
    forward slashes.  While backward slashes will allow correctly locating the
    filesystem resource locally, it most likely will result in incorrect
    destinations on the remote side, as backward slashes will be interpreted as
    part of the name itself.
    """
    def __init__(self, path):
        """Path is any (absolute or relative) Windows path."""
        if os.path.isabs(path):
            drive, tail = os.path.splitdrive(path)
            self.drive = drive.rstrip(':')
            self.tail = tail.lstrip('\\').lstrip('/')
        else:
            self.drive = ''
            self.tail = path.lstrip('\\').lstrip('/')

    def __repr__(self):
        """This is where the rsync safe conversion takes place.  If there is a
        space in the path, encapsulates the path with quotes to assure the path
        will not be misinterpreted as being two or more paths.
        """
        drive = os.path.join('\\cygdrive', self.drive) if self.drive else ''
        path = os.path.join(drive, self.tail)
        path = path.replace('\\', '/')
        if " " in path:
            path = '"%s"' % path
        return path

def fix_path(path):
    return str(WindowsPath(path)) if plat == 'Windows' else path

_remote_path_delimiter = ':'

def is_local_path(path):
    return ((_remote_path_delimiter not in path) or
            (plat == 'Windows' and re.match(r'[a-zA-Z]:(\\|/)', path)))

def parse_local_paths(local_paths):
    """Validate local paths."""
    if not isinstance(local_paths, (tuple, list)):
        local_paths = [local_paths]

    parsed_paths = []
    for path in local_paths:
        path = fix_path(path)
        if path.count(_remote_path_delimiter):
            raise cloud.CloudException('Local path cannot contain "%s"' %
                                       _remote_path_delimiter)
        parsed_paths.append(path)
    return parsed_paths

def parse_remote_paths(remote_paths):
    """Validate remote paths."""
    if not isinstance(remote_paths, (tuple, list)):
        remote_paths = [remote_paths]

    resource_name = None
    parsed_paths = []
    for path in remote_paths:
        if path.count(_remote_path_delimiter) != 1:
            raise cloud.CloudException('Remote path must be resource_name:path')
        r_name, r_path = path.split(_remote_path_delimiter)
        resource_name = resource_name or r_name
        if not r_name:
            raise cloud.CloudException('Remote path must start with a resource name')
        if resource_name != r_name:
            raise cloud.CloudException('All remote paths must use the same resource')
        parsed_paths.append(r_path)
    return (resource_name, parsed_paths)


########## OS interaction ###########
def exec_command(cmd, pipe_output=True, redirect_stderr=False):
    """Simple shell exec wrapper, with no stdin."""
    out_pipe = PIPE if pipe_output else None
    err_pipe = STDOUT if redirect_stderr else out_pipe
    p = Popen(cmd, shell=True, stdout=out_pipe, stderr=err_pipe)
    stdout, stderr = p.communicate()
    status = p.returncode
    return (status, stdout, stderr)

def ssh_session(username, hostname, key_path, port=None, run_cmd=None):
    default_options = ['-o', 'StrictHostKeyChecking=no',
                       '-o', 'ConnectTimeout=60',
                       '-o', 'LogLevel=QUIET',
                       ]
    ssh_path = _ssh_path()
    cmd = [ssh_path]
    if port:
        cmd.extend(['-p', str(port)])
    cmd.extend(default_options)
    cmd.extend(['-i', fix_path(key_path)])
    cmd.append('%s@%s' % (username, hostname))
    if run_cmd:
        if hasattr(run_cmd, '__iter__'):
            run_cmd = ' '.join(run_cmd)
        cmd.append("'%s'" % run_cmd)
        pipe_output = True
    else:
        pipe_output = False
    cmd_str = ' '.join(cmd)
    return exec_command(cmd_str, pipe_output=pipe_output, redirect_stderr=True)

def rsync_session(src_arg, dest_arg, delete=False, pipe_output=True):
    """Perform an rsync operation over ssh using api_key's ssh key."""
    _check_rsync_dependencies()

    api_key = cloud.connection_info().get('api_key')
    key_file = credentials.get_sshkey_path(api_key)
    if not os.path.exists(key_file):
        raise cloud.CloudException('%s not present. Something must have errored '
                                   'earlier. See cloud.log' % key_file)
    key_file = str(WindowsPath(key_file)) if plat == 'Windows' else key_file

    ssh_shell = ' '.join([_ssh_path(), '-q',
                          '-o UserKnownHostsFile=/dev/null',
                          '-o StrictHostKeyChecking=no',
                          '-o LogLevel=QUIET',
                          '-i %s' % key_file])
    rsync_cmd = [_rsync_path(), '-avz']
    if plat == 'Windows':
        rsync_cmd.append('--chmod=u+rwx')
    if delete:
        rsync_cmd.append('--delete')
    rsync_cmd.extend(['-e "%s"' % ssh_shell, src_arg, dest_arg])
    rsync_cmd_str = ' '.join(rsync_cmd)
    cloudLog.debug('Performing rsync command: %s', rsync_cmd_str)
    print_stdout('Performing rsync...')
    return exec_command(rsync_cmd_str, pipe_output)

def ssh_server_job(keep_alive=False):
    """A job that waits until all ssh connections have terminated to close
    Set *keep_alive* to True to keep job running forever
    """
    poll_time = 4.0

    activated = False

    if keep_alive: # sleep forever
        time.sleep(315360000)
        return

    while True:
        time.sleep(poll_time)
        p = Popen('ps -C sshd -o cmd | grep pts', shell=True, stdout=PIPE, stderr=PIPE)
        stdout, _ = p.communicate()
        lines = stdout.splitlines()

        if not lines:
            if activated:
                print 'No more SSH connections. Terminating job'
                break

        else:
            activated=True


